import json
import logging
import os
from pathlib import Path

import click
from flask import Flask, current_app, jsonify, request
from waitress import serve

from seagoat import __version__
from seagoat.cache import Cache, get_cache_root
from seagoat.queue.task_queue import TaskQueue
from seagoat.utils.config import GLOBAL_CONFIG_FILE, get_config_values
from seagoat.utils.server import (
    ServerDoesNotExist,
    get_free_port,
    get_server_info,
    get_servers_info,
    is_server_running,
    stop_server,
    update_server_info,
)
from seagoat.utils.wait import wait_for


def create_app(repo_path):
    logging.info("Creating server...")
    app = Flask(__name__)
    app.config["PROPAGATE_EXCEPTIONS"] = True

    app.extensions["task_queue"] = TaskQueue(
        repo_path=repo_path, minimum_chunks_to_analyze=0
    )

    @app.route("/query/<query>")
    def query_codebase(query):
        limit_clue = request.args.get("limitClue", "500")
        context_above = request.args.get("contextAbove", 3)
        context_below = request.args.get("contextBelow", 3)

        try:
            limit_clue = int(limit_clue)
        except ValueError as exception:
            raise RuntimeError(
                "Invalid limitClue value. Must be an integer."
            ) from exception

        result = current_app.extensions["task_queue"].enqueue(
            "query",
            query=query,
            context_below=int(context_below),
            context_above=int(context_above),
            limit_clue=limit_clue,
        )

        return result

    @app.route("/status")
    def status_():
        stats = current_app.extensions["task_queue"].enqueue(
            "get_stats",
        )
        return {
            "stats": stats,
            "version": __version__,
        }

    @app.errorhandler(Exception)
    def handle_exception(exception_to_handle):
        res = {
            "code": 500,
            "error": {
                "type": "Internal Server Error",
                "message": exception_to_handle.message
                if hasattr(exception_to_handle, "message")
                else f"{exception_to_handle}",
            },
        }

        return jsonify(res), 500

    return app


def start_server(repo_path, custom_port=None):
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    app = create_app(repo_path)
    port = custom_port

    if port is None:
        port = get_free_port()

    update_server_info(
        repo_path,
        {
            "host": "localhost",
            "port": port,
            "address": f"http://localhost:{port}",
            "pid": os.getpid(),
        },
    )
    serve(app, host="0.0.0.0", port=port, threads=1)


def get_server(repo_path, custom_port=None):
    port = None
    try:
        server_info = get_server_info(repo_path)
        server_address = server_info["address"]

        if is_server_running(repo_path):
            click.echo(f"Server is already running at {server_address}")
            return server_address
    except ServerDoesNotExist:
        pass

    if custom_port is not None:
        port = custom_port

    os.environ["TOKENIZERS_PARALLELISM"] = "true"

    start_server(str(repo_path), custom_port=port)

    wait_for(lambda: get_server_info(repo_path), timeout=60)

    server_info = get_server_info(repo_path)
    click.echo(f"Server started at {server_info['address']}")
    return server_info["address"]


@click.group()
@click.version_option(version=__version__, prog_name="seagoat")
def server():
    """
    This server analyzes your codebase and creates vector embeddings for it.

    You can query this server using the seagoat command.
    """
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")


@server.command()
@click.argument("repo_path")
@click.option("--port", type=int, help="The port to start the server on", default=None)
def start(repo_path, port):
    """Starts the server."""
    config = get_config_values(Path(repo_path))
    port = port if port is not None else config["server"]["port"]

    get_server(repo_path, custom_port=port)
    click.echo("Server running.")


def get_status_data(repo_path):
    """Return the status data of the server."""
    status_info = {"isRunning": False, "url": None, "pid": None}

    try:
        server_info = get_server_info(repo_path)
        server_address = server_info["address"]
        pid = server_info.get("pid", None)

        if is_server_running(repo_path):
            status_info = {
                "isRunning": True,
                "url": server_address,
                "pid": pid,
            }

    except ServerDoesNotExist:
        pass

    return status_info


@server.command()
@click.argument("repo_path")
@click.option("--json", "use_json_format", is_flag=True, help="Output status as JSON.")
def status(repo_path, use_json_format):
    """Checks the status of the server."""
    status_info = get_status_data(repo_path)

    if use_json_format:
        click.echo(json.dumps(status_info))
    else:
        if status_info["isRunning"]:
            click.echo(f"Server is running at {status_info['url']}")
        else:
            click.echo("Server is not running.")


@server.command()
@click.argument("repo_path")
def stop(repo_path):
    """Stops the server."""
    try:
        stop_server(repo_path)
    except ServerDoesNotExist:
        click.echo(
            f"No server information found for {repo_path}. It might not be running or was never started."
        )
        return

    click.echo(
        "Server stopped. If it was running, it will stop after finishing current tasks."
    )


@server.command(name="server-info")
def _server_info():
    """Returns information about all SeaGOAT servers in JSON format."""
    servers_info = get_servers_info()
    formatted_servers_info = {}

    for normalized_repo_path, info in servers_info.items():
        formatted_servers_info[normalized_repo_path] = {
            "repoPath": normalized_repo_path,
            "cacheLocation": {
                "chroma": str(
                    Cache("chroma", Path(normalized_repo_path), {}).get_cache_folder(),
                ),
                "ripgrep": str(
                    Cache("ripgrep", Path(normalized_repo_path), {}).get_cache_folder(),
                ),
            },
            "isRunning": is_server_running(normalized_repo_path),
            "host": info["host"],
            "port": info["port"],
            "address": f"http://{info['host']}:{info['port']}",
        }

    info = {
        "version": __version__,
        "servers": formatted_servers_info,
        "globalCache": str(get_cache_root()),
        "globalConfigFile": str(GLOBAL_CONFIG_FILE),
    }

    click.echo(json.dumps(info))


if __name__ == "__main__":
    server()



import os
import socket
from pathlib import Path
from typing import Dict, TypedDict, Union

import appdirs
import psutil

from seagoat.utils.json_file import get_json_file_contents, write_to_json_file


class ServerInfo(TypedDict):
    address: str
    host: str
    port: int
    pid: int


class ServerDoesNotExist(Exception):
    pass


def _get_server_data_file_path() -> Path:
    user_cache_dir = Path(appdirs.user_cache_dir("seagoat-servers"))
    user_cache_dir.mkdir(parents=True, exist_ok=True)
    return user_cache_dir / "serverData.json"


def normalize_repo_path(repo_path: Union[str, Path]) -> str:
    return str(os.path.normpath(Path(repo_path).expanduser().resolve()))


def get_servers_info() -> Dict[str, ServerInfo]:
    path = _get_server_data_file_path()
    if not os.path.exists(path):
        write_to_json_file(path, {})

    contents = get_json_file_contents(path)
    if contents is None:
        contents = {}

    for key in list(contents.keys()):
        if not os.path.exists(key):
            del contents[key]

    return contents


def update_server_info(
    repo_path: Union[str, Path], new_server_data: ServerInfo
) -> None:
    servers_info = get_servers_info()
    repo_id = normalize_repo_path(repo_path)
    servers_info[repo_id] = new_server_data

    write_to_json_file(_get_server_data_file_path(), servers_info)


def stop_server(repo_path: Union[str, Path]) -> None:
    servers_info = get_servers_info()
    repo_id = normalize_repo_path(repo_path)
    if repo_id in servers_info:
        server_info = servers_info[repo_id]
        process = psutil.Process(server_info["pid"])

        servers_info.pop(repo_id)
        write_to_json_file(_get_server_data_file_path(), servers_info)

        process.terminate()
        process.wait()
    else:
        raise ServerDoesNotExist(f"Server for {repo_path} does not exist.")


def get_server_info(repo_path: Union[str, Path]) -> ServerInfo:
    servers_info = get_servers_info()
    repo_id = normalize_repo_path(repo_path)

    if repo_id not in servers_info:
        raise ServerDoesNotExist(f"Server for {repo_path} does not exist.")

    server_info = servers_info[repo_id]
    return server_info


def is_server_running(repo_path: Union[str, Path]) -> bool:
    try:
        server_info = get_server_info(repo_path)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as socket_obj:
            return (
                socket_obj.connect_ex((server_info["host"], server_info["port"])) == 0
            )
    except ServerDoesNotExist:
        return False


def get_free_port() -> int:
    socket_obj = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_obj.bind(("", 0))
    _, port = socket_obj.getsockname()
    socket_obj.close()

    return port



import functools
import re
from collections import Counter
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, List, Set

from seagoat.utils.file_reader import read_file_with_correct_encoding
from seagoat.utils.file_types import get_file_penalty_factor

SPLITTER_PATTERN = re.compile(r"\s+")


@functools.lru_cache(maxsize=100)
def compile_regex_pattern(query: str):
    terms = re.split(SPLITTER_PATTERN, query)
    pattern = ".*".join(map(re.escape, terms))

    return re.compile(pattern, re.IGNORECASE)


@functools.lru_cache(maxsize=10_000)
def get_number_of_exact_matches(line: str, query: str):
    pattern = compile_regex_pattern(query)

    if re.search(pattern, line):
        return 1
    return 0


def get_best_score(result) -> float:
    best_score = min(
        (x for x in result.lines.values() if ResultLineType.RESULT in x.types),
        key=lambda item: item.get_score(),
    ).get_score()

    best_score *= get_file_penalty_factor(result.full_path)

    return best_score


class ResultLineType(Enum):
    RESULT = "result"
    CONTEXT = "context"
    BRIDGE = "bridge"

    def __str__(self):
        return self.value


@dataclass(frozen=True)
class ResultLine:
    parent: "Result"
    line: int
    vector_distance: float
    line_text: str
    types: Set[ResultLineType]

    def get_score(self) -> float:
        return self.vector_distance / (
            1 + get_number_of_exact_matches(self.line_text, self.parent.query_text)
        )

    def add_type(self, type_: ResultLineType) -> None:
        self.types.add(type_)

    def to_json(self):
        return {
            "score": round(self.get_score(), 4),
            "line": self.line,
            "lineText": self.line_text,
            "resultTypes": list(sorted(set(str(t) for t in self.types))),
        }


@dataclass(frozen=True)
class ResultBlock:
    lines: List[ResultLine]

    def _get_line_count_per_type(self) -> Dict[str, int]:
        counts = Counter()

        for line in self.lines:
            for type_ in line.types:
                counts[str(type_)] += 1

        return dict(counts)

    def _get_score(self) -> float:
        return min(
            line.get_score()
            for line in self.lines
            if ResultLineType.RESULT in line.types
        )

    def to_json(self):
        return {
            "score": self._get_score(),
            "lines": [line.to_json() for line in self.lines],
            "lineTypeCount": self._get_line_count_per_type(),
        }


class Result:
    def __init__(self, query_text: str, path: str, full_path: Path) -> None:
        self.path: str = path
        self.query_text: str = query_text
        self.full_path: Path = full_path
        self.lines: Dict[int, ResultLine] = {}
        self.line_texts = read_file_with_correct_encoding(self.full_path).splitlines()

    def __repr__(self) -> str:
        return f"Result(path={self.path})"

    def extend(self, other) -> None:
        self.lines.update(other.lines)

    def add_line(self, line: int, vector_distance: float) -> None:
        types = set()
        if line in self.lines and self.lines[line].vector_distance < vector_distance:
            vector_distance = self.lines[line].vector_distance
            types = self.lines[line].types

        types.add(
            ResultLineType.RESULT,
        )

        self.lines[line] = ResultLine(
            self,
            line,
            vector_distance,
            self.line_texts[line - 1],
            types,
        )

    def get_lines(self):
        best_score = get_best_score(self)

        return list(
            sorted(
                set(
                    result_line.line
                    for result_line in self.lines.values()
                    if (result_line.get_score() <= best_score * 10)
                    or ResultLineType.CONTEXT in result_line.types
                )
            )
        )

    def _merge_almost_touching_blocks(self, blocks):
        if len(blocks) == 0:
            return []

        merged_blocks = [blocks[0]]

        for block in blocks[1:]:
            last_block = merged_blocks[-1]
            last_line_of_last_block = last_block.lines[-1].line
            line_range_from_last_block = range(
                last_line_of_last_block + 1, block.lines[0].line
            )

            if len(line_range_from_last_block) <= 2:
                for line in line_range_from_last_block:
                    bridge_line = ResultLine(
                        self,
                        line,
                        0.0,
                        self.line_texts[line - 1],
                        {ResultLineType.BRIDGE},
                    )
                    last_block.lines.append(bridge_line)

                for line in block.lines:
                    last_block.lines.append(line)
            else:
                merged_blocks.append(block)

        return merged_blocks

    def get_result_blocks(self):
        self_lines = self.get_lines()
        lines_to_include = [
            line
            for line in sorted(self.lines.values(), key=lambda item: item.line)
            if line.line in self_lines
        ]
        blocks = []

        for line in lines_to_include:
            distance_from_previous_line = (
                line.line - blocks[-1].lines[-1].line if blocks else 0
            )

            if not blocks or distance_from_previous_line > 1:
                blocks.append(ResultBlock(lines=[]))

            blocks[-1].lines.append(line)

        return self._merge_almost_touching_blocks(blocks)

    def to_json(self):
        return {
            "path": self.path,
            "fullPath": str(self.full_path),
            "score": round(get_best_score(self), 4),
            "blocks": [block.to_json() for block in self.get_result_blocks()],
        }

    def add_context_lines(self, lines: int):
        if lines == 0:
            return

        direction = lines // abs(lines)
        for result_line, _ in list(self.lines.items()):
            for offset in range(abs(lines)):
                new_line = result_line + (offset + 1) * direction

                if (new_line) not in range(len(self.line_texts)):
                    continue

                if new_line not in self.lines:
                    self.lines[new_line] = ResultLine(
                        self,
                        line=new_line,
                        vector_distance=0.0,
                        line_text=self.line_texts[new_line - 1],
                        types={ResultLineType.CONTEXT},
                    )

                self.lines[new_line].add_type(ResultLineType.CONTEXT)



from chardet.universaldetector import UniversalDetector


def read_file_with_correct_encoding(file_path):
    try:
        # Try to read the file as UTF-8 first
        with open(file_path, encoding="utf-8") as file:
            return file.read()
    except UnicodeDecodeError:
        # If UTF-8 reading fails, then detect encoding and read accordingly
        detector = UniversalDetector()
        with open(file_path, "rb") as file:
            for line in file:
                detector.feed(line)
                if detector.done:
                    break
        detector.close()
        encoding = detector.result["encoding"] or "utf-8"

        with open(file_path, encoding=encoding) as file:
            return file.read()



from pathlib import Path
from typing import Union

IGNORED_BRANCHES = {"gh-pages"}

TEXT_FILE_TYPES = {
    ".txt",
    ".md",
}

SUPPORTED_FILE_TYPES = TEXT_FILE_TYPES | {
    ".py",
    ".c",
    ".cc",
    ".cpp",
    ".cxx",
    ".h",
    ".hpp",
    ".ts",
    ".js",
    ".tsx",
    ".jsx",
    ".html",
    ".go",
    ".java",
    ".php",
    ".rb",
}


def is_file_type_supported(path: Union[Path, str]):
    return Path(path).suffix in SUPPORTED_FILE_TYPES


def get_file_penalty_factor(path: Union[Path, str]) -> float:
    # Text file lines are penalized compared to code file lines as they
    # generally have more meaningful words, but the user might be more likely
    # to be looking for code than text.
    if Path(path).suffix in TEXT_FILE_TYPES:
        return 1.5

    return 1.0



from pathlib import Path

import chromadb
from chromadb.config import Settings
from chromadb.errors import IDAlreadyExistsError
from chromadb.utils import embedding_functions

from seagoat.cache import Cache
from seagoat.repository import Repository
from seagoat.result import Result
from seagoat.utils.config import get_config_values

MAXIMUM_VECTOR_DISTANCE = 1.5


def get_metadata_and_distance_from_chromadb_result(chromadb_results):
    return (
        list(
            zip(
                chromadb_results["metadatas"][0],
                chromadb_results["distances"][0],
            )
        )
        if chromadb_results["metadatas"] and chromadb_results["distances"]
        else None
    ) or []


def format_results(query_text: str, repository, chromadb_results):
    files = {}

    for metadata, distance in get_metadata_and_distance_from_chromadb_result(
        chromadb_results
    ):
        if distance > MAXIMUM_VECTOR_DISTANCE:
            break
        path = str(metadata["path"])
        line = int(metadata["line"])
        full_path = Path(repository.path) / path

        if not full_path.exists():
            continue

        if path not in files:
            files[path] = Result(query_text, path, full_path)
        files[path].add_line(line, distance)

    return files.values()


def initialize(repository: Repository):
    cache = Cache("chroma", Path(repository.path), {})
    config = get_config_values(Path(repository.path))

    chroma_client = chromadb.PersistentClient(
        path=str(cache.get_cache_folder()),
        settings=Settings(
            anonymized_telemetry=False,
        ),
    )
    embedding_function_name = config["server"]["chroma"]["embeddingFunction"]["name"]
    embedding_function_kwargs = config["server"]["chroma"]["embeddingFunction"][
        "arguments"
    ]
    embedding_function = getattr(embedding_functions, embedding_function_name)(
        **embedding_function_kwargs
    )
    chroma_collection = chroma_client.get_or_create_collection(
        name="code_data", embedding_function=embedding_function
    )

    def fetch(query_text: str, limit: int):
        # Slightly overfetch results as it will sorted using a different score later
        maximum_chunks_to_fetch = 100  # this should be plenty, especially because many times context could be included
        n_results = min((limit + 1) * 2, maximum_chunks_to_fetch)

        chromadb_results = chroma_collection.query(
            query_texts=[query_text],
            n_results=n_results,
        )

        return format_results(query_text, repository, chromadb_results)

    def cache_chunk(chunk):
        try:
            chroma_collection.add(
                ids=[chunk.chunk_id],
                documents=[chunk.chunk],
                metadatas=[{"path": chunk.path, "line": chunk.codeline}],
            )
        except IDAlreadyExistsError:
            pass

    def cache_repo():
        # chromadb does not need any repo cache action
        pass

    return {
        "fetch": fetch,
        "cache_chunk": cache_chunk,
        "cache_repo": cache_repo,
    }



import logging
import mmap
import platform
import re
import subprocess
import tempfile
from pathlib import Path

from stop_words import get_stop_words

from seagoat.cache import Cache
from seagoat.repository import Repository
from seagoat.result import Result
from seagoat.sources.chroma import MAXIMUM_VECTOR_DISTANCE
from seagoat.utils.file_reader import read_file_with_correct_encoding
from seagoat.utils.file_types import is_file_type_supported

KILOBYTE = 1024
MEGABYTE = KILOBYTE * 1024
MAX_MMAP_SIZE = 100
MAX_MMAP_SIZE_BYTES = MAX_MMAP_SIZE * MEGABYTE
MAX_FILE_SIZE = 200 * KILOBYTE
STOP_WORDS = set(get_stop_words("english"))


class RipGrepCache(str):
    def __init__(self, repository):
        cache = Cache("ripgrep", Path(repository.path), {})
        self.repository = repository
        self.file_path = cache.get_cache_folder() / "mmap"
        if platform.system() == "Windows":
            self.file_path = tempfile.mktemp()
        self.is_initialized = False
        self._data = ""

    def _iterate_files_to_cache(self):
        for file, _ in self.repository.top_files():
            yield file

    def _iterate_lines_to_cache(self):
        for file in self._iterate_files_to_cache():
            file_contents = read_file_with_correct_encoding(file.absolute_path)

            if len(file_contents) > MAX_FILE_SIZE:
                logging.warning("Warning: file %s is too large to cache", file.path)
                continue

            for line_number, line in enumerate(file_contents.splitlines(), start=1):
                yield file, line_number, line

    def _generate_cache_lines(self):
        for file, line_number, line in self._iterate_lines_to_cache():
            yield f"{file.path}:{line_number}:{line}\n"

    def _build_cache_file(self):
        total_estimated_cache_size = 0
        line_count = 0

        with open(self.file_path, "w", encoding="utf-8") as cache_file:
            for formattted_cache_line in self._generate_cache_lines():
                cache_file.write(formattted_cache_line)
                total_estimated_cache_size += len(formattted_cache_line)
                line_count += 1

                if total_estimated_cache_size > MAX_MMAP_SIZE_BYTES:
                    logging.warning(
                        "Warning: maximum estimated ripgrep cache size of %s megabytes exceeded",
                        MAX_MMAP_SIZE,
                    )

                    break

        logging.info(
            "Estimated ripgrep cache size: %.2f megabytes (%s bytes). Line count %s",
            total_estimated_cache_size / MEGABYTE,
            int(total_estimated_cache_size),
            line_count,
        )
        self.is_initialized = True

    def rebuild(self):
        self._build_cache_file()
        if platform.system() == "Windows":
            # Memory map does not work on Windows for some reason
            # Use a simple string as a fallback
            with open(self.file_path, encoding="utf-8") as cache_file:
                self._data = cache_file.read()
        else:
            with open(self.file_path, "r+b") as cache_file:
                self._data = mmap.mmap(cache_file.fileno(), 0)

    def encode(self, *args, **kwargs):
        return self._data

    def as_input(self):
        if platform.system() == "Windows":
            return self.encode()

        return self


def _fetch(query_text: str, path: str, limit: int, cache: RipGrepCache):
    query_text_without_stopwords = " ".join(
        query for query in query_text.split(" ") if query not in STOP_WORDS
    )
    if len(query_text_without_stopwords) > 2:
        query_text = query_text_without_stopwords
    query_text = re.sub(r"\s+", "|", query_text)
    files = {}

    cmd = [
        "rg",
        "--max-count",
        str(limit),
        "--ignore-case",
        query_text,
    ]

    try:
        rg_output = subprocess.check_output(
            cmd, encoding="utf-8", input=cache.as_input()
        )
    except subprocess.CalledProcessError as exception:
        rg_output = exception.output
    for line in rg_output.splitlines():
        relative_path, raw_line_number, _ = line.split(":", 2)
        line_number = int(raw_line_number)

        absolute_path = Path(path) / relative_path

        if not is_file_type_supported(relative_path):
            continue

        if relative_path not in files:
            files[relative_path] = Result(query_text, str(relative_path), absolute_path)

        # This is so that ripgrep results are on comparable levels with chroma results
        files[relative_path].add_line(line_number, MAXIMUM_VECTOR_DISTANCE * 0.8)

    return files.values()


def initialize(repository: Repository):
    path = repository.path
    memory_cache = RipGrepCache(repository)

    def cache_chunk(_):
        # Ripgrep does not use chunks
        pass

    def cache_repo():
        memory_cache.rebuild()

    def fetch(query_text: str, limit: int):
        if not memory_cache.is_initialized:
            repository.analyze_files()
            cache_repo()

        return _fetch(query_text, str(path), limit, memory_cache)

    return {
        "fetch": fetch,
        "cache_chunk": cache_chunk,
        "cache_repo": cache_repo,
    }
